import { ClientBase, ClientConfig } from 'pg';
import { ServiceConfig } from './local-replication-service';
import { InboxMessage } from './models';
/** The inbox service configuration */
export type InboxServiceConfig = ServiceConfig & {
    /**
     * Database connection details. The user needs update permission to the inbox.
     */
    pgConfig: ClientConfig;
    settings: ServiceConfig['settings'] & {
        /**
         * The maximum number of retries to handle an incoming inbox message. Defaults to 5.
         */
        maxRetries?: number;
    };
};
/**
 * Message handler for a specific aggregate type and event type.
 */
export interface InboxMessageHandler {
    /** The aggregate root type */
    aggregateType: string;
    /** The name of the event created for the aggregate type. */
    eventType: string;
    /**
     * Custom business logic to handle a message that was stored in the inbox.
     * @param message The inbox message with the payload to handle.
     * @param client The database client that is part of a transaction to safely handle the inbox message.
     * @throws If something failed and the inbox message should NOT be acknowledged - throw an error.
     */
    handle: (message: InboxMessage, client: ClientBase) => Promise<void>;
}
/**
 * Initialize the service to watch for inbox table inserts.
 * @param config The configuration object with required values to connect to the WAL.
 * @param messageHandlers A list of message handlers to handle the inbox messages. I
 * @returns Functions for a clean shutdown and to help testing "outages" of the inbox service
 */
export declare const initializeInboxService: (config: InboxServiceConfig, messageHandlers: InboxMessageHandler[]) => Promise<[shutdown: () => Promise<void>]>;
//# sourceMappingURL=inbox-service.d.ts.map