"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.initializeInboxService = void 0;
const pg_1 = require("pg");
const inbox_1 = require("./inbox");
const local_replication_service_1 = require("./local-replication-service");
const logger_1 = require("./logger");
const utils_1 = require("./utils");
/**
 * Initialize the service to watch for inbox table inserts.
 * @param config The configuration object with required values to connect to the WAL.
 * @param messageHandlers A list of message handlers to handle the inbox messages. I
 * @returns Functions for a clean shutdown and to help testing "outages" of the inbox service
 */
const initializeInboxService = (config, messageHandlers) => __awaiter(void 0, void 0, void 0, function* () {
    const pool = createPgPool(config);
    const messageHandler = createMessageHandler(messageHandlers, pool, config);
    const errorResolver = createErrorResolver(pool, config);
    const [shutdown] = yield (0, local_replication_service_1.createService)(config, messageHandler, errorResolver, mapInboxRetries);
    return [
        () => __awaiter(void 0, void 0, void 0, function* () {
            pool.removeAllListeners();
            pool
                .end()
                .catch((e) => (0, logger_1.logger)().error(e, 'PostgreSQL pool shutdown error'));
            shutdown().catch((e) => (0, logger_1.logger)().error(e, 'Inbox service shutdown error'));
        }),
    ];
});
exports.initializeInboxService = initializeInboxService;
const createPgPool = (config) => {
    const pool = new pg_1.Pool(config.pgConfig);
    pool.on('error', (err) => {
        (0, logger_1.logger)().error(err, 'PostgreSQL pool error');
    });
    return pool;
};
/**
 * Executes the inbox verification, the actual message handler, and marks the
 * inbox message as processed in one transaction.
 */
const createMessageHandler = (messageHandlers, pool, config) => {
    return (message) => __awaiter(void 0, void 0, void 0, function* () {
        yield (0, utils_1.executeTransaction)(pool, (client) => __awaiter(void 0, void 0, void 0, function* () {
            const result = yield (0, inbox_1.verifyInbox)(message, client, config);
            if (result === true) {
                yield Promise.all(messageHandlers
                    .filter(({ aggregateType, eventType }) => aggregateType === message.aggregateType &&
                    eventType === message.eventType)
                    .map(({ handle }) => handle(message, client)));
                yield (0, inbox_1.ackInbox)(message, client, config);
            }
            else {
                (0, logger_1.logger)().warn(message, `Received inbox message cannot be processed: ${result}`);
            }
        }));
    });
};
/**
 * Handle specific error cases (message already processed/not found) by
 * acknowledging the inbox WAL message. For other errors: increase the retry
 * counter of the message and retry it later.
 */
const createErrorResolver = (pool, config) => {
    /**
     * An error handler that will increase the inbox retries count on transient errors.
     * It returns true if the message should nevertheless be acknowledged.
     * @returns true to retry the message - otherwise false
     */
    return (message, _error) => __awaiter(void 0, void 0, void 0, function* () {
        try {
            return yield (0, utils_1.executeTransaction)(pool, (client) => __awaiter(void 0, void 0, void 0, function* () {
                const action = yield (0, inbox_1.nackInbox)(message, client, config);
                if (action === 'RETRIES_EXCEEDED') {
                    return true; // giving up - acknowledge the WAL message
                }
                else {
                    return false; // retry WAL message
                }
            }));
        }
        catch (error) {
            (0, logger_1.logger)().error(Object.assign(Object.assign({}, message), { err: error }), 'The message handling error handling failed.');
            return false; // retry WAL message
        }
    });
};
/** The local replication service maps by default only the outbox properties */
const mapInboxRetries = (input) => {
    var _a;
    if ('retries' in input &&
        typeof input.retries === 'number' &&
        'processed_at' in input &&
        (input.processed_at == null || input.processed_at instanceof Date)) {
        return {
            retries: input.retries,
            processedAt: (_a = input.processed_at) === null || _a === void 0 ? void 0 : _a.toISOString(),
        };
    }
    return {};
};
//# sourceMappingURL=inbox-service.js.map