"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
/* eslint-disable @typescript-eslint/no-non-null-assertion */
const inspector_1 = __importDefault(require("inspector"));
const inbox_1 = require("./inbox");
const logger_1 = require("./logger");
const pg_1 = require("pg");
const isDebugMode = () => inspector_1.default.url() !== undefined;
if (isDebugMode()) {
    jest.setTimeout(600000);
}
else {
    (0, logger_1.disableLogger)(); // Hide logs if the tests are not run in debug mode
    jest.setTimeout(7000);
}
const message = {
    id: 'message_id',
    aggregateType: 'test_type',
    eventType: 'test_event_type',
    aggregateId: 'test_aggregate_id',
    payload: { result: 'success' },
    createdAt: '2023-01-18T21:02:27.000Z',
};
const inboxMessage = Object.assign(Object.assign({}, message), { retries: 0, processedAt: null });
const config = {
    pgConfig: {
        host: 'test_host',
        port: 5432,
        database: 'test_db',
        user: 'test_user',
        password: 'test_user_password',
    },
    pgReplicationConfig: {
        host: 'test_host',
        port: 5432,
        database: 'test_db',
        user: 'test_inbox_user',
        password: 'test_inbox_user_password',
    },
    settings: {
        dbSchema: 'test_schema',
        dbTable: 'test_table',
        postgresPub: 'test_pub',
        postgresSlot: 'test_slot',
        maxRetries: 7,
    },
};
// required for the initializeInboxMessageStorage
jest.mock('pg', () => {
    return Object.assign(Object.assign({}, jest.requireActual('pg')), { Pool: jest.fn().mockImplementation(() => ({
            connect: jest.fn(() => new pg_1.Client()),
            on: jest.fn(),
            end: jest.fn(() => Promise.resolve()),
            removeAllListeners: jest.fn(),
        })), Client: jest.fn().mockImplementation(() => ({
            query: jest.fn((sql, _params) => __awaiter(void 0, void 0, void 0, function* () {
                if (sql.indexOf(`INSERT INTO ${config.settings.dbSchema}.${config.settings.dbTable}`)) {
                    return { rowCount: 1 };
                }
                throw new Error(`Missed to mock the following SQL query: ${sql}`);
            })),
            on: jest.fn(),
            release: jest.fn(),
        })) });
});
jest.mock('./utils', () => {
    return Object.assign(Object.assign({}, jest.requireActual('./utils')), { executeTransaction: jest.fn((pool, callback) => __awaiter(void 0, void 0, void 0, function* () {
            const client = yield pool.connect();
            const response = yield callback(client);
            client.release();
            return response;
        })) });
});
describe('Inbox unit tests', () => {
    describe('initializeInboxMessageStorage', () => {
        test('it initializes the inbox message storage and stores a message without an error', () => __awaiter(void 0, void 0, void 0, function* () {
            // Act
            const [storeInboxMessage, shutdown] = yield (0, inbox_1.initializeInboxMessageStorage)(config);
            // Assert
            yield expect(storeInboxMessage(message)).resolves.not.toThrow();
            yield shutdown();
        }));
    });
    describe('verifyInbox', () => {
        test('it verifies the inbox message', () => __awaiter(void 0, void 0, void 0, function* () {
            // Arrange
            const pool = new pg_1.Pool();
            const client = yield pool.connect();
            // Act + Assert
            // Test for INBOX_MESSAGE_NOT_FOUND (no row found)
            client.query = jest.fn().mockResolvedValue({
                rowCount: 0,
            });
            let result = yield (0, inbox_1.verifyInbox)(inboxMessage, client, config);
            expect(result).toBe('INBOX_MESSAGE_NOT_FOUND');
            // Test for ALREADY_PROCESSED (one row found but has a processed date)
            client.query = jest.fn().mockResolvedValue({
                rowCount: 1,
                rows: [{ processed_at: new Date() }],
            });
            result = yield (0, inbox_1.verifyInbox)(inboxMessage, client, config);
            expect(result).toBe('ALREADY_PROCESSED');
            // Test for success (one row found that was not processed yet)
            client.query = jest.fn().mockResolvedValue({
                rowCount: 1,
                rows: [{ processed_at: null }],
            });
            result = yield (0, inbox_1.verifyInbox)(inboxMessage, client, config);
            expect(result).toBe(true);
        }));
    });
    describe('ackInbox', () => {
        it('should call query with the correct parameters', () => __awaiter(void 0, void 0, void 0, function* () {
            // Arrange
            const pool = new pg_1.Pool();
            const client = yield pool.connect();
            client.query = jest.fn().mockResolvedValue({
                rowCount: 0,
            });
            // Act
            yield (0, inbox_1.ackInbox)(inboxMessage, client, config);
            // Assert
            expect(client.query).toHaveBeenCalledWith(`UPDATE ${config.settings.dbSchema}.${config.settings.dbTable} SET processed_at = $1 WHERE id = $2`, [expect.any(String), inboxMessage.id]);
        }));
    });
    describe('nackInbox', () => {
        it.each([1, 2, 3, 4, 5, 6])('should return RETRY when retries are less than maxRetries: %p', (retries) => __awaiter(void 0, void 0, void 0, function* () {
            // Arrange
            const pool = new pg_1.Pool();
            const client = yield pool.connect();
            client.query = jest.fn().mockResolvedValue({
                rowCount: 1,
                rows: [{ retries }], // query response --> after the increment
            });
            // Act
            const result = yield (0, inbox_1.nackInbox)(Object.assign(Object.assign({}, inboxMessage), { retries: retries - 1 }), client, config);
            // Assert
            expect(client.query).toHaveBeenCalledWith(expect.stringContaining('UPDATE test_schema.test_table SET retries = retries + 1 WHERE id = $1'), [inboxMessage.id]);
            expect(result).toBe('RETRY');
        }));
        it.each([7, 8, 999])('should return RETRIES_EXCEEDED when retries are equal or larger than maxRetries: %p', (retries) => __awaiter(void 0, void 0, void 0, function* () {
            // Arrange
            const pool = new pg_1.Pool();
            const client = yield pool.connect();
            client.query = jest.fn().mockResolvedValue({
                rowCount: 1,
                rows: [{ retries }], // query response --> after the increment
            });
            // Act
            const result = yield (0, inbox_1.nackInbox)(Object.assign(Object.assign({}, inboxMessage), { retries: retries - 1 }), client, config);
            // Assert
            expect(client.query).toHaveBeenCalledWith(expect.stringContaining('UPDATE test_schema.test_table SET retries = retries + 1 WHERE id = $1'), [inboxMessage.id]);
            expect(result).toBe('RETRIES_EXCEEDED');
        }));
        it('should return RETRIES_EXCEEDED when the corresponding inbox row was not found', () => __awaiter(void 0, void 0, void 0, function* () {
            // Arrange
            const pool = new pg_1.Pool();
            const client = yield pool.connect();
            client.query = jest.fn().mockResolvedValue({
                rowCount: 0,
                rows: [],
            });
            // Act
            const result = yield (0, inbox_1.nackInbox)(Object.assign({}, inboxMessage), client, config);
            // Assert
            expect(client.query).toHaveBeenCalledWith(expect.stringContaining('UPDATE test_schema.test_table SET retries = retries + 1 WHERE id = $1'), [inboxMessage.id]);
            expect(result).toBe('RETRIES_EXCEEDED');
        }));
    });
});
//# sourceMappingURL=inbox.spec.js.map