"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.__only_for_unit_tests__ = exports.createService = void 0;
const pg_logical_replication_1 = require("pg-logical-replication");
const logger_1 = require("./logger");
const utils_1 = require("./utils");
/**
 * Initiate the outbox/inbox table to listen for WAL messages.
 * @param config The replication connection settings and general service settings
 * @param messageHandler The message handler that handles the outbox/inbox message
 * @param errorHandler A handler that can decide if the WAL message should be acknowledged (true) or not (restarts the logical replication service)
 * @param mapAdditionalRows The inbox table requires an additional row to be mapped to the inbox message
 * @returns A function to stop the service
 */
const createService = ({ pgReplicationConfig, settings }, messageHandler, errorHandler, mapAdditionalRows) => __awaiter(void 0, void 0, void 0, function* () {
    const plugin = new pg_logical_replication_1.PgoutputPlugin({
        protoVersion: 1,
        publicationNames: [settings.postgresPub],
    });
    let service;
    let stopped = false;
    // Run the service in an endless background loop until it gets stopped
    (() => __awaiter(void 0, void 0, void 0, function* () {
        var _a;
        while (!stopped) {
            try {
                yield new Promise((resolve, reject) => {
                    let heartbeatAckTimer = undefined;
                    service = new pg_logical_replication_1.LogicalReplicationService(pgReplicationConfig, {
                        acknowledge: { auto: false, timeoutSeconds: 0 },
                    });
                    service.on('data', (lsn, log) => __awaiter(void 0, void 0, void 0, function* () {
                        const msg = getRelevantMessage(log, settings, mapAdditionalRows);
                        if (msg) {
                            // 'OutboxMessage' is assignable to the constraint of type 'T',
                            // but 'T' could be instantiated with a different subtype of constraint 'OutboxMessage'
                            const message = msg;
                            (0, logger_1.logger)().trace(message, `Received a WAL message for ${settings.dbSchema}.${settings.dbTable}`);
                            try {
                                yield messageHandler(message);
                                clearTimeout(heartbeatAckTimer);
                                yield service.acknowledge(lsn);
                            }
                            catch (error) {
                                const err = (0, utils_1.ensureError)(error);
                                (0, logger_1.logger)().error(Object.assign(Object.assign({}, message), { err }), err.message);
                                if (errorHandler) {
                                    const acknowledge = yield errorHandler(message, err);
                                    if (acknowledge && !service.isStop()) {
                                        yield service.acknowledge(lsn);
                                    }
                                    else if (!service.isStop()) {
                                        service.emit('error', error);
                                    }
                                }
                                else if (!service.isStop()) {
                                    service.emit('error', error);
                                }
                            }
                        }
                    }));
                    service.on('error', (err) => __awaiter(void 0, void 0, void 0, function* () {
                        service.removeAllListeners();
                        yield service.stop();
                        reject(err);
                    }));
                    service.on('heartbeat', (lsn, _timestamp, shouldRespond) => __awaiter(void 0, void 0, void 0, function* () {
                        if (shouldRespond) {
                            heartbeatAckTimer = setTimeout(() => __awaiter(void 0, void 0, void 0, function* () {
                                (0, logger_1.logger)().trace(`${lsn}: acknowledged heartbeat`);
                                yield service.acknowledge(lsn);
                            }), 5000);
                        }
                    }));
                    service
                        .subscribe(plugin, settings.postgresSlot)
                        .then(() => resolve(true))
                        .catch((err) => __awaiter(void 0, void 0, void 0, function* () {
                        (0, logger_1.logger)().error({ err }, 'Logical replication subscription error');
                        service.removeAllListeners();
                        yield service.stop();
                        reject(err);
                    }));
                });
            }
            catch (err) {
                yield (0, utils_1.sleep)((_a = settings.restartDelay) !== null && _a !== void 0 ? _a : 1000);
                (0, logger_1.logger)().error({ err }, 'LogicalReplicationService error');
            }
        }
    }))();
    return [
        () => __awaiter(void 0, void 0, void 0, function* () {
            (0, logger_1.logger)().error('started cleanup');
            stopped = true;
            service === null || service === void 0 ? void 0 : service.removeAllListeners();
            service === null || service === void 0 ? void 0 : service.stop().catch((e) => (0, logger_1.logger)().error(e, 'Error on service shutdown.'));
        }),
    ];
});
exports.createService = createService;
/** Get and map the inbox/outbox message if the WAL log event is such an event. Otherwise returns undefined. */
const getRelevantMessage = (log, { dbSchema, dbTable }, mapAdditionalRows) => log.tag === 'insert' &&
    log.relation.schema === dbSchema &&
    log.relation.name === dbTable
    ? mapMessage(log.new, mapAdditionalRows)
    : undefined;
/** Maps the WAL log entry to an outbox or inbox message */
const mapMessage = (input, mapAdditionalRows) => {
    if (typeof input !== 'object' || input === null) {
        return undefined;
    }
    if (!('id' in input) ||
        typeof input.id !== 'string' ||
        !('aggregate_type' in input) ||
        typeof input.aggregate_type !== 'string' ||
        !('aggregate_id' in input) ||
        typeof input.aggregate_id !== 'string' ||
        !('event_type' in input) ||
        typeof input.event_type !== 'string' ||
        !('created_at' in input) ||
        !(input.created_at instanceof Date) || // date
        !('payload' in input)) {
        return undefined;
    }
    const additional = mapAdditionalRows === null || mapAdditionalRows === void 0 ? void 0 : mapAdditionalRows(input);
    const message = Object.assign({ id: input.id, aggregateType: input.aggregate_type, aggregateId: input.aggregate_id, eventType: input.event_type, payload: input.payload, createdAt: input.created_at.toISOString() }, additional);
    return message;
};
/**
 * This export is _only_ done for unit tests as the createService function is
 * otherwise very hard to unit test. Exports work only for jest tests!
 */
exports.__only_for_unit_tests__ = {};
if (process.env.JEST_WORKER_ID) {
    exports.__only_for_unit_tests__.getRelevantMessage = getRelevantMessage;
    exports.__only_for_unit_tests__.mapMessage = mapMessage;
}
//# sourceMappingURL=local-replication-service.js.map