"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
/* eslint-disable @typescript-eslint/no-explicit-any */
/* eslint-disable @typescript-eslint/no-non-null-assertion */
const inspector_1 = __importDefault(require("inspector"));
const outbox_service_1 = require("./outbox-service");
const logger_1 = require("./logger");
const isDebugMode = () => inspector_1.default.url() !== undefined;
if (isDebugMode()) {
    jest.setTimeout(600000);
}
else {
    (0, logger_1.disableLogger)(); // Hide logs if the tests are not run in debug mode
    jest.setTimeout(7000);
}
const repService = {};
jest.mock('pg-logical-replication', () => {
    return Object.assign(Object.assign({}, jest.requireActual('pg-logical-replication')), { LogicalReplicationService: jest.fn().mockImplementation(() => {
            const lrs = {
                handleData: undefined,
                handleError: undefined,
                on: (event, listener) => {
                    switch (event) {
                        case 'data':
                            repService.handleData = listener;
                            break;
                        case 'error':
                            repService.handleError = listener;
                            break;
                    }
                },
                acknowledge: jest.fn(),
                removeAllListeners: jest.fn(),
                emit: jest.fn(),
                stop: jest.fn(() => Promise.resolve()),
                subscribe: () => new Promise(() => {
                    /** never return */
                }),
                isStop: () => true,
            };
            repService.acknowledge = lrs.acknowledge;
            repService.stop = lrs.stop;
            return lrs;
        }) });
});
const outboxDbMessage = {
    id: 'message_id',
    aggregate_type: 'test_type',
    event_type: 'test_event_type',
    aggregate_id: 'test_aggregate_id',
    payload: { result: 'success' },
    created_at: new Date('2023-01-18T21:02:27.000Z'),
};
const outboxMessage = {
    id: outboxDbMessage.id,
    aggregateType: outboxDbMessage.aggregate_type,
    aggregateId: outboxDbMessage.aggregate_id,
    eventType: outboxDbMessage.event_type,
    payload: outboxDbMessage.payload,
    createdAt: '2023-01-18T21:02:27.000Z',
};
const config = {
    pgReplicationConfig: {
        host: 'test_host',
        port: 5432,
        database: 'test_db',
        user: 'test_outbox_user',
        password: 'test_outbox_user_password',
    },
    settings: {
        dbSchema: 'test_schema',
        dbTable: 'test_table',
        postgresPub: 'test_pub',
        postgresSlot: 'test_slot',
    },
};
const relation = {
    tag: 'relation',
    relationOid: 1,
    schema: 'test_schema',
    name: 'test_table',
    replicaIdentity: 'default',
    columns: [
        {
            name: 'id',
            flags: 0,
            typeOid: 23,
            typeMod: -1,
            typeSchema: 'pg_catalog',
            typeName: 'int4',
            parser: (raw) => raw,
        },
    ],
    keyColumns: ['id'],
};
describe('Outbox service unit tests - initializeOutboxService', () => {
    beforeEach(() => {
        repService.handleData = undefined;
        repService.handleError = undefined;
        repService.acknowledge = undefined;
        repService.stop = undefined;
    });
    it('should call the messageHandler and acknowledge the WAL message when no errors are thrown', () => __awaiter(void 0, void 0, void 0, function* () {
        // Arrange
        const messageHandler = jest.fn(() => Promise.resolve());
        const [cleanup] = yield (0, outbox_service_1.initializeOutboxService)(config, messageHandler);
        expect(repService.handleData).toBeDefined();
        expect(repService.handleError).toBeDefined();
        // Act
        yield repService.handleData('0/00000001', {
            tag: 'insert',
            relation,
            new: outboxDbMessage,
        });
        // Assert
        expect(messageHandler).toHaveBeenCalledWith(outboxMessage);
        expect(repService.acknowledge).toHaveBeenCalledWith('0/00000001');
        expect(repService.stop).not.toHaveBeenCalled();
        yield cleanup();
        expect(repService.stop).toHaveBeenCalledTimes(1);
    }));
    it('should call the messageHandler but not acknowledge the WAL message when the message handler throws an error', () => __awaiter(void 0, void 0, void 0, function* () {
        // Arrange
        const throwErrorMessageHandler = jest.fn(() => __awaiter(void 0, void 0, void 0, function* () {
            throw new Error('Unit Test');
        }));
        const [cleanup] = yield (0, outbox_service_1.initializeOutboxService)(config, throwErrorMessageHandler);
        expect(repService.handleData).toBeDefined();
        expect(repService.handleError).toBeDefined();
        // Act
        yield repService.handleData('0/00000001', {
            tag: 'insert',
            relation,
            new: outboxDbMessage,
        });
        // Assert
        expect(throwErrorMessageHandler).toHaveBeenCalledWith(outboxMessage);
        expect(repService.acknowledge).not.toHaveBeenCalled();
        expect(repService.stop).not.toHaveBeenCalled();
        yield cleanup();
        expect(repService.stop).toHaveBeenCalledTimes(1);
    }));
});
//# sourceMappingURL=outbox-service.spec.js.map