"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const inspector_1 = __importDefault(require("inspector"));
const logger_1 = require("./logger");
const utils_1 = require("./utils");
const isDebugMode = () => inspector_1.default.url() !== undefined;
if (isDebugMode()) {
    jest.setTimeout(600000);
}
else {
    (0, logger_1.disableLogger)(); // Hide logs if the tests are not run in debug mode
    jest.setTimeout(7000);
}
describe('Utils Unit Tests', () => {
    describe('ensureError', () => {
        it('should return the input error if it is an instance of Error', () => {
            // Arrange
            const inputError = new Error('This is an error');
            // Act
            const error = (0, utils_1.ensureError)(inputError);
            // Assert
            expect(error).toBe(inputError);
        });
        it('should return a new Error if the input is not an instance of Error', () => {
            // Arrange
            const input = 'This is not an error';
            // Act
            const error = (0, utils_1.ensureError)(input);
            // Assert
            expect(error).toEqual(new Error(input));
        });
    });
    describe('sleep', () => {
        it('should sleep for the given amount of milliseconds', () => __awaiter(void 0, void 0, void 0, function* () {
            // Arrange
            const sleepTime = 1000;
            const startTime = Date.now();
            // Act
            yield (0, utils_1.sleep)(sleepTime);
            // Assert
            const endTime = Date.now();
            expect(endTime - startTime).toBeGreaterThanOrEqual(sleepTime);
        }));
    });
    describe('executeTransaction', () => {
        let pool;
        let callback;
        beforeEach(() => {
            pool = {
                connect: jest.fn().mockResolvedValue({
                    on: jest.fn(),
                    query: jest.fn(),
                    release: jest.fn(),
                    listeners: jest.fn().mockReturnValue([]),
                }),
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
            };
            callback = jest.fn().mockResolvedValue('SELECT 1;');
        });
        afterEach(() => {
            jest.clearAllMocks();
        });
        it('should open a transaction and execute the callback', () => __awaiter(void 0, void 0, void 0, function* () {
            // Act
            yield (0, utils_1.executeTransaction)(pool, callback);
            // Assert
            expect(pool.connect).toHaveBeenCalled();
            expect(callback).toHaveBeenCalled();
        }));
        it('should return the result of the callback', () => __awaiter(void 0, void 0, void 0, function* () {
            // Act
            const result = yield (0, utils_1.executeTransaction)(pool, callback);
            // Assert
            expect(result).toBe('SELECT 1;');
        }));
        it('should commit the transaction if the callback resolves', () => __awaiter(void 0, void 0, void 0, function* () {
            // Arrange
            const client = yield pool.connect();
            // Act
            yield (0, utils_1.executeTransaction)(pool, callback);
            // Assert
            expect(client.query).toHaveBeenCalledWith('COMMIT');
            expect(client.release).toHaveBeenCalled();
        }));
        it('should rollback the transaction if the callback throws', () => __awaiter(void 0, void 0, void 0, function* () {
            // Arrange
            callback.mockRejectedValue(new Error('Callback error'));
            const client = yield pool.connect();
            // Act + Assert
            yield expect((0, utils_1.executeTransaction)(pool, callback)).rejects.toThrow();
            expect(client.query).toHaveBeenCalledWith('ROLLBACK');
            expect(client.release).toHaveBeenCalled();
        }));
    });
});
//# sourceMappingURL=utils.spec.js.map